/* 
 * tkMacWinMenu.c --
 *
 *	This module implements the common elements of the Mac and Windows
 *	specific features of menus. This file is not used for UNIX.
 *
 * Copyright (c) 1996 by Sun Microsystems, Inc.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 * SCCS: @(#) tkMacWinMenu.c 1.34 96/12/09 16:55:41
 */

#include "tkMenu.h"

static int postCommandGeneration;

static int			PreprocessMenu _ANSI_ARGS_((Menu *menuPtr));


/*
 *----------------------------------------------------------------------
 *
 * PreprocessMenu --
 *
 *	The guts of the preprocessing. Recursive.
 *
 * Results:
 *	The return value is a standard Tcl result (errors can occur
 *	while the postcommands are being processed).
 *
 * Side effects:
 *	Since commands can get executed while this routine is being executed,
 *	the entire world can change.
 *
 *----------------------------------------------------------------------
 */

static int
PreprocessMenu(menuPtr)
    Menu *menuPtr;
{
    int index, result, finished;
    Menu *cascadeMenuPtr;
   
    Tcl_Preserve((ClientData) menuPtr);
    
    /*
     * First, let's process the post command on ourselves. If this command
     * destroys this menu, or if there was an error, we are done.
     */
     
    result = TkPostCommand(menuPtr);
    if ((result != TCL_OK) || (menuPtr->tkwin == NULL)) {
    	goto done;
    }
    

    /*
     * Build up the platform structure if it is not there yet.
     */
     
    TkMenuCompleteIdleHandlers(menuPtr);

    /*
     * Now, we go through structure and process all of the commands.
     * Since the structure is changing, we stop after we do one command,
     * and start over. When we get through without doing any, we are done.
     */
    
    
    do {
    	finished = 1;
        for (index = 0; index < menuPtr->numEntries; index++) {
            if ((menuPtr->entries[index]->type == CASCADE_ENTRY)
            	    && (menuPtr->entries[index]->name != NULL)) {
            	if ((menuPtr->entries[index]->childMenuRefPtr != NULL)
            		&& (menuPtr->entries[index]->childMenuRefPtr->menuPtr
            		!= NULL)) {
            	    cascadeMenuPtr =
            	    	    menuPtr->entries[index]->childMenuRefPtr->menuPtr;
            	    if (cascadeMenuPtr->postCommandGeneration != 
            	    	    postCommandGeneration) {
            	    	cascadeMenuPtr->postCommandGeneration = 
            	    		postCommandGeneration;
            	        result = PreprocessMenu(cascadeMenuPtr);
            	        if (result != TCL_OK) {
            	            goto done;
            	        }
            	        finished = 0;
            	        break;
            	    }
            	}
            }
        }
    } while (!finished);
    
    done:
    Tcl_Release((ClientData)menuPtr);
    return result;
}

/*
 *----------------------------------------------------------------------
 *
 * TkPreprocessMenu --
 *
 *	On the Mac and on Windows, all of the postcommand processing has
 *	to be done on the entire tree underneath the main window to be
 *	posted. This means that we have to traverse the menu tree and
 *	issue the postcommands for all of the menus that have cascades
 *	attached. Since the postcommands can change the menu structure while
 *	we are traversing, we have to be extremely careful. Basically, the
 *	idea is to traverse the structure until we succesfully process
 *	one postcommand. Then we start over, and do it again until
 *	we traverse the whole structure without processing any postcommands.
 *
 *	We are also going to set up the cascade back pointers in here
 *	since we have to traverse the entire structure underneath the menu
 *	anyway, We can clear the postcommand marks while we do that.
 *
 * Results:
 *	The return value is a standard Tcl result (errors can occur
 *	while the postcommands are being processed).
 *
 * Side effects:
 *	Since commands can get executed while this routine is being executed,
 *	the entire world can change.
 *
 *----------------------------------------------------------------------
 */

int
TkPreprocessMenu(menuPtr)
    Menu *menuPtr;
{
    postCommandGeneration++;
    menuPtr->postCommandGeneration = postCommandGeneration;
    return PreprocessMenu(menuPtr);
}

/*
 *----------------------------------------------------------------------
 *
 * 
 * TkpChangeMenuType --
 *
 *	If the menu is actually changing type, we have to reallocate the
 *	data structures.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Allocations are done and the old platform is freed.
 *
 *----------------------------------------------------------------------
 */

void
TkpChangeMenuType(menuPtr)
    Menu *menuPtr;
{

    if ((menuPtr->menuType == TEAROFF_MENU) && (menuPtr->transient)) {
	Tcl_DoWhenIdle(TkGetPlatformData, (ClientData) menuPtr);
	menuPtr->menuFlags |= PLATFORM_DATA_PENDING;
	menuPtr->menuType = MASTER_MENU;
    } else if ((menuPtr->menuType == MASTER_MENU) && !(menuPtr->transient)) {
    	int i;
    	
	if (menuPtr->menuFlags & PLATFORM_DATA_PENDING) {
	    Tcl_CancelIdleCall(TkGetPlatformData, (ClientData) menuPtr);
	    menuPtr->menuFlags &= PLATFORM_DATA_PENDING;
	}
	if (menuPtr->platformData != NULL) {
    	    for (i = menuPtr->numEntries - 1; i >= 0; i--) {
    		TkpDestroyMenuEntry(menuPtr->entries[i]);
    	    }
    	    TkpDestroyMenu(menuPtr);
	}
	Tcl_DoWhenIdle(TkGetPlatformData, (ClientData) menuPtr);
	menuPtr->menuFlags |= PLATFORM_DATA_PENDING;
    	menuPtr->menuType = TEAROFF_MENU;
    }
}


/*
 *----------------------------------------------------------------------
 *
 * TkpDrawMenuEntryBackground --
 *
 *	This procedure draws the background part of a menu.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Commands are output to X to display the menu in its
 *	current mode.
 *
 *----------------------------------------------------------------------
 */

void
TkpDrawMenuEntryBackground(
    Menu *menuPtr,			/* The menu we are drawing. */
    MenuEntry *mePtr,			/* The entry we are drawing. */
    Drawable d,				/* What we are drawing into */
    Tk_3DBorder activeBorder,		/* Border for active items */
    Tk_3DBorder bgBorder,		/* Border for the background */
    int x,				/* left edge */
    int y,				/* top edge */
    int width,				/* width of rectangle to draw */
    int height)				/* height of rectangle to draw */
{
    if (mePtr->state == tkActiveUid) {
	bgBorder = activeBorder;
    }
    Tk_Fill3DRectangle(menuPtr->tkwin, d, bgBorder,
    	    x, y, width, height, 0, TK_RELIEF_FLAT);
}
