/* 
 * tkMacMenu.c --
 *
 *	This module implements the Mac-platform specific features of menus.
 *
 * Copyright (c) 1996 by Sun Microsystems, Inc.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 * SCCS: @(#) tkMacMenu.c 1.65 96/12/18 17:02:48
 */

#include <Limits.h>
#include <Menus.h>
#include <OSUtils.h>
#include <Palettes.h>
#include <Resources.h>
#include <string.h>
#include <ToolUtils.h>
#include "tkMenu.h"
#include "tkMacInt.h"
#include "tkMenuButton.h"

/*
 * Various geometry definitions:
 */

#define CASCADE_ARROW_HEIGHT 	10
#define CASCADE_ARROW_WIDTH 	8
#define DECORATION_BORDER_WIDTH 2
#define MARGIN_WIDTH 		8

/*
 * The following are constants relating to the SICNs used for drawing the MDEF.
 */

#define SICN_RESOURCE_NUMBER		128

#define SICN_HEIGHT 		16
#define SICN_ROWS 		2
#define CASCADE_ICON_WIDTH	7
#define	SHIFT_ICON_WIDTH	10
#define	OPTION_ICON_WIDTH	16
#define CONTROL_ICON_WIDTH	12
#define COMMAND_ICON_WIDTH	10

#define CASCADE_ARROW		0
#define SHIFT_ICON		1
#define OPTION_ICON		2
#define CONTROL_ICON		3
#define COMMAND_ICON		4

/*
 * Platform specific flags for menu entries
 *
 * ENTRY_COMMAND_ACCEL		Indicates the entry has the command key
 *				in its accelerator string
 * ENTRY_OPTION_ACCEL		Indicates the entry has the option key
 *				in its accelerator string
 * ENTRY_SHIFT_ACCEL		Indicates the entry has the shift key
 *				in its accelerator string
 * ENTRY_CONTROL_ACCEL		Indicates the entry has the control key
 *				in its accelerator string
 */

#define ENTRY_COMMAND_ACCEL	ENTRY_PLATFORM_FLAG1
#define ENTRY_OPTION_ACCEL	ENTRY_PLATFORM_FLAG2
#define ENTRY_SHIFT_ACCEL	ENTRY_PLATFORM_FLAG3
#define ENTRY_CONTROL_ACCEL	ENTRY_PLATFORM_FLAG4
#define ENTRY_ACCEL_MASK	(ENTRY_COMMAND_ACCEL | ENTRY_OPTION_ACCEL \
				| ENTRY_SHIFT_ACCEL | ENTRY_CONTROL_ACCEL)

#define CASCADE_CMD (0x1b)    	
				/* The special command char for cascade
			         * menus. */
#define SEPARATOR_TEXT "\p(-"
				/* The text for a menu separator. */

#define MENUBAR_REDRAW_PENDING 1

static Tcl_HashTable commandTable;
				/* The list of menuInstancePtrs associated with
				 * menu ids */
static char *currentMenuBarName;
				/* Malloced. Name of current menu in menu bar.
				 * NULL if no menu set. TO DO: make this a
				 * DString. */
static Tk_Window currentMenuBarOwner;
				/* Which window owns the current menu bar. */
static int inPostMenu;		/* We cannot be re-entrant like X
				 * windows. */
static short lastMenuID;	/* To pass to NewMenu; need to figure out
				 * a good way to do this. */
static unsigned char lastCascadeID;
				/* Cascades have to have ids that are
				 * less than 256. */
static MacDrawable macMDEFDrawable;
				/* Drawable for use by MDEF code */
static int menuBarFlags;	/* Used for whether the menu bar needs
				 * redrawing or not. */
static TkMenuDefUPP menuDefProc;/* The routine descriptor to the MDEF proc.
				 * The MDEF is needed to draw menus with
				 * non-standard attributes and to support
				 * tearoff menus. */
static struct TearoffSelect {
    Menu *menuPtr;		/* The menu that is torn off */
    Point point;		/* The point to place the new menu */
    Rect excludeRect;		/* We don't want to drag tearoff highlights
    				 * when we are in this menu */
} tearoffStruct;

/*
 * Forward declarations for procedures defined later in this file:
 */
 
static int		GetNewID _ANSI_ARGS_((Tcl_Interp *interp,
			    Menu *menuInstPtr, int cascade, 
			    short *menuIDPtr));
static char		FindMarkCharacter _ANSI_ARGS_((MenuEntry *mePtr));
static void		FreeID _ANSI_ARGS_((int menuID));
static void		MenuDefProc _ANSI_ARGS_((short message,
			    MenuHandle menu, Rect *menuRectPtr,
			    Point hitPt, short *whichItem,
			    TkMenuLowMemGlobals *globalsPtr));
static void		RecursivelyDeleteMenu _ANSI_ARGS_((
			    Menu *menuPtr));
static void		RecursivelyInsertMenu _ANSI_ARGS_((
			    Menu *menuPtr));
static void		SetDefaultMenubar _ANSI_ARGS_((void));
static void		SetMenuCascade _ANSI_ARGS_((Menu *menuPtr));
static void		SetMenuIndicator _ANSI_ARGS_((MenuEntry *mePtr));


/*
 *----------------------------------------------------------------------
 *
 * TkMacUseID --
 *
 *	Take the ID out of the available list for new menus. Used by the
 *	default menu bar's menus so that they do not get created at the tk
 *	level. See GetNewID for more information.
 *
 * Results:
 *	Returns TCL_OK if the id was not in use. Returns TCL_ERROR if the
 *	id was in use.
 *
 * Side effects:
 *	A hash table entry in the command table is created with a NULL
 *	value.
 *
 *----------------------------------------------------------------------
 */

int
TkMacUseMenuID(
    int macID)			/* The id to take out of the table */
{
    Tcl_HashEntry *commandEntryPtr;
    int newEntry;
    
    TkMenuInit();
    commandEntryPtr = Tcl_CreateHashEntry(&commandTable, (char *) macID,
    	    &newEntry);
    if (newEntry == 1) {
    	Tcl_SetHashValue(commandEntryPtr, NULL);
    	return TCL_OK;
    } else {
    	return TCL_ERROR;
    }
}

/*
 *----------------------------------------------------------------------
 *
 * GetNewID --
 *
 *	Allocates a new menu id and marks it in use. Each menu on the
 *	mac must be designated by a unique id, which is a short. In
 *	addition, some ids are reserved by the system. Since Tk uses
 *	mostly dynamic menus, we must allocate and free these ids on
 *	the fly. We use the id as a key into a hash table; if there
 *	is no hash entry, we know that we can use the id.
 *
 * Results:
 *	Returns TCL_OK if succesful; TCL_ERROR if there are no more
 *	ids of the appropriate type to allocate. menuIDPtr contains
 *	the new id if succesful.
 *
 * Side effects:
 *	An entry is created for the menu in the command hash table,
 *	and the hash entry is stored in the appropriate field in the
 *	menu data structure.
 *
 *----------------------------------------------------------------------
 */

static int
GetNewID(
    Tcl_Interp *interp,		/* Used for error reporting */
    Menu *menuPtr,		/* The menu we are working with */
    int cascade,		/* 0 if we are working with a normal menu;
    				   1 if we are working with a cascade */
    short *menuIDPtr)		/* The resulting id */
{
    int found = 0;
    int newEntry;
    Tcl_HashEntry *commandEntryPtr;
    short returnID = *menuIDPtr;

    /*
     * The following code relies on shorts and unsigned chars wrapping
     * when the highest value is incremented. Also, the values between
     * 236 and 255 inclusive are reserved for DA's by the Mac OS.
     */
    
    if (!cascade) {
    	short curID = lastMenuID + 1;
        if (curID == 236) {
    	    curID = 256;
    	}

    	while (curID != lastMenuID) {
    	    commandEntryPtr = Tcl_CreateHashEntry(&commandTable,
		    (char *) curID, &newEntry);
    	    if (newEntry == 1) {
    	        found = 1;
    	        lastMenuID = returnID = curID;
    	        break;
    	    }
    	    curID++;
    	    if (curID == 236) {
    	    	curID = 256;
    	    }
    	}
    } else {
    
    	/*
    	 * Cascade ids must be between 0 and 235 only, so they must be
    	 * dealt with separately.
    	 */
    
    	unsigned char curID = lastCascadeID + 1;
        if (curID == 236) {
    	    curID = 0;
    	}
    	
    	while (curID != lastCascadeID) {
    	    commandEntryPtr = Tcl_CreateHashEntry(&commandTable,
		    (char *) curID, &newEntry);
    	    if (newEntry == 1) {
    	    	found = 1;
    	    	lastCascadeID = returnID = curID;
    	    	break;
    	    }
    	    curID++;
    	    if (curID == 236) {
    	    	curID = 0;
    	    }
    	}
    }

    if (found) {
    	Tcl_SetHashValue(commandEntryPtr, (char *) menuPtr);
    	*menuIDPtr = returnID;
    	return TCL_OK;
    } else {
        Tcl_AppendResult(interp, "No more menus can be allocated.", 
        	(char *) NULL);
    	return TCL_ERROR;
    }
}

/*
 *----------------------------------------------------------------------
 *
 * FreeID --
 *
 *	Marks the id as free.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	The hash table entry for the ID is cleared.
 *
 *----------------------------------------------------------------------
 */

static void
FreeID(
    int menuID)			/* The id to free */
{
    Tcl_HashEntry *entryPtr = Tcl_FindHashEntry(&commandTable,
	    (char *) menuID);
    
    if (entryPtr != NULL) {
    	 Tcl_DeleteHashEntry(entryPtr);
    }
}

/*
 *----------------------------------------------------------------------
 *
 * TkpNewMenu --
 *
 *	Gets a new blank menu. Only the platform specific options are filled
 *	in.
 *
 * Results:
 *	Returns a pointer to the menu structure. Not sure what to do when
 *	we can't get a Mac menu handle.
 *
 * Side effects:
 *	Allocates a Macintosh menu handle.
 *
 *----------------------------------------------------------------------
 */

PlatformData
TkpNewMenu(
    Menu *menuPtr)		/* The common structure we are making the
    				 * platform structure for. */
{
    short menuID;
    Str255 itemText;
    int length;
    MenuHandle macMenuHdl;
    
    if (GetNewID(menuPtr->interp, menuPtr, 0, &menuID) != TCL_OK) {
    	macMenuHdl = NULL;
    } else {
    	length = strlen(Tk_PathName(menuPtr->tkwin));
    	memmove(&itemText[1], Tk_PathName(menuPtr->tkwin), 
    		(length > 230) ? 230 : length);
    	itemText[0] = (length > 230) ? 230 : length;
    	macMenuHdl = NewMenu(menuID, itemText);
#ifdef GENERATINGCFM
	{
	    Handle mdefProc = GetResource('MDEF', 591);
	    Handle sicnHandle = GetResource('SICN', SICN_RESOURCE_NUMBER);
	    if ((mdefProc != NULL) && (sicnHandle != NULL)) {
    	    	(*macMenuHdl)->menuProc = mdefProc;
    	    }
    	}
#endif
    }
    
    return (PlatformData) macMenuHdl;
}

/*
 *----------------------------------------------------------------------
 *
 * TkpDestroyMenu --
 *
 *	Destroys platform-specific menu structures.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	All platform-specific allocations are freed up.
 *
 *----------------------------------------------------------------------
 */

void
TkpDestroyMenu(
    Menu *menuPtr)		/* The common menu structure */
{
    MenuHandle macMenuHdl = (MenuHandle) menuPtr->platformData;

    DeleteMenu((*macMenuHdl)->menuID);
    FreeID((*macMenuHdl)->menuID);
    DisposeMenu(macMenuHdl);
    menuPtr->platformData = NULL;
}

/*
 *----------------------------------------------------------------------
 *
 * SetMenuCascade --
 *
 *	Does any cleanup to change a menu from a normal to a cascade.
 *
 * Results:
 *	None
 *
 * Side effects:
 *	The mac menu id is reset.
 *
 *----------------------------------------------------------------------
 */

static void
SetMenuCascade(
    Menu* menuPtr)		/* The menu we are setting up to be a
				 * cascade. */
{
    MenuHandle macMenuHdl = (MenuHandle) menuPtr->platformData;
    short newMenuID, menuID = (*macMenuHdl)->menuID;
    
    if (menuID >= 256) {
    	if (GetNewID(menuPtr->interp, menuPtr, 1, &newMenuID) == TCL_OK) {
    	    FreeID(menuID);
    	    (*macMenuHdl)->menuID = newMenuID;
    	}
    }
}

/*
 *----------------------------------------------------------------------
 *
 * TkpDestroyMenuEntry --
 *
 *	Cleans up platform-specific menu entry items.
 *
 * Results:
 *	None
 *
 * Side effects:
 *	All platform-specific allocations are freed up.
 *
 *----------------------------------------------------------------------
 */

void
TkpDestroyMenuEntry(
    MenuEntry *mePtr)		/* The common structure for the menu 
    				 * entry. */
{
    MenuHandle macMenuHdl = (MenuHandle) mePtr->menuPtr->platformData;
  
    DeleteMenuItem(macMenuHdl, mePtr->index + 1);
}

/*
 *----------------------------------------------------------------------
 *
 * GetEntryText --
 *
 *	Given a menu entry, gives back the text that should go in it.
 *	Separators should be done by the caller, as they have to be
 *	handled specially.
 *
 * Results:
 *	itemText points to the new text for the item.
 *
 * Side effects:
 *	None.
 *
 *----------------------------------------------------------------------
 */

static void
GetEntryText(
    MenuEntry *mePtr,		/* A pointer to the menu entry. */
    Str255 itemText)		/* The pascal string containing the text */
{
    if (mePtr->type == TEAROFF_ENTRY) {
	strcpy((char *)itemText, (const char *)"\p(Tear-off)");
    } else if (mePtr->imageString != NULL) {
	strcpy((char *)itemText, (const char *)"\p(Image)");
    } else if (mePtr->bitmap != None) {
	strcpy((char *)itemText, (const char *)"\p(Pixmap)");
    } else if (mePtr->label == NULL || mePtr->labelLength == 0) {
    
	/*
	 * The Mac menu manager does not like null strings.
	 */
	 
	strcpy((char *)itemText, (const char *)"\p ");
    } else {
	memmove(&itemText[1], mePtr->label, 
	    	(mePtr->labelLength > 230 ? 230 : mePtr->labelLength));
	itemText[0] = mePtr->labelLength > mePtr->labelLength 
	    	? 230 : mePtr->labelLength;
    }
}

/*
 *----------------------------------------------------------------------
 *
 * FindMarkCharacter --
 *
 *	Finds the Macintosh mark character based on the font of the
 *	item. We calculate a good mark character based on the font
 * 	that this item is rendered in.
 *
 * 	We try the following special mac characters. If none of them
 * 	are present, just use the check mark.
 * 	'' - Check mark character
 * 	'' - Bullet character
 * 	'' - Filled diamond
 * 	'' - Hollow diamond
 * 	'' = Long dash ("em dash")
 * 	'-' = short dash (minus, "en dash");
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	New item is added to platform menu
 *
 *----------------------------------------------------------------------
 */

static char
FindMarkCharacter(
    MenuEntry *mePtr)		/* The entry we are finding the character
    				 * for. */
{
    char markChar;
    Tk_Font tkfont = (mePtr->tkfont == NULL) ? mePtr->menuPtr->tkfont
    	    : mePtr->tkfont;
    	    
    if (!TkMacIsCharacterMissing(tkfont, '')) {
    	markChar = '';
    } else if (!TkMacIsCharacterMissing(tkfont, '')) {
    	markChar = '';
    } else if (!TkMacIsCharacterMissing(tkfont, '')) {
    	markChar = '';
    } else if (!TkMacIsCharacterMissing(tkfont, '')) {
    	markChar = '';
    } else if (!TkMacIsCharacterMissing(tkfont, '')) {
    	markChar = '';
    } else if (!TkMacIsCharacterMissing(tkfont, '-')) {
    	markChar = '-';
    } else {
    	markChar = '';
    }
    return markChar;
}

/*
 *----------------------------------------------------------------------
 *
 * SetMenuIndicator --
 *
 *	Sets the Macintosh mark character based on the font of the
 *	item.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	New item is added to platform menu
 *
 *----------------------------------------------------------------------
 */

static void
SetMenuIndicator(
    MenuEntry *mePtr)		/* The entry we are setting */
{
    Menu *menuPtr = mePtr->menuPtr;
    MenuHandle macMenuHdl = (MenuHandle) menuPtr->platformData;
    char markChar;

    /*
     * There can be no indicators on menus that are not checkbuttons
     * or radiobuttons. However, we should go ahead and set them
     * so that menus look right when they are displayed. We should
     * not set cascade entries, however, as the mark character
     * means something different for cascade items on the Mac.
     * Also, we do reflect the tearOff menu items in the Mac menu
     * handle, so we ignore them.
     */

    if ((mePtr->type == CASCADE_ENTRY) || (mePtr->type == TEAROFF_MENU)) {
    	return;
    }
     
    if (((mePtr->type == RADIO_BUTTON_ENTRY) 
    	    || (mePtr->type == CHECK_BUTTON_ENTRY))
    	    && (mePtr->entryFlags & ENTRY_SELECTED)) {
    	markChar = FindMarkCharacter(mePtr);
    } else {
        markChar = 0;
    }
    SetItemMark(macMenuHdl, mePtr->index + 1, markChar);
}

/*
 *----------------------------------------------------------------------
 *
 * SetMenuTitle --
 *
 *	Sets title of menu so that the text displays correctly in menubar.
 *	This code directly manipulates menu handle data. This code
 *	was originally part of an ancient Apple Developer Response mail.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	The menu handle will change size depending on the length of the
 *	title
 *
 *----------------------------------------------------------------------
 */

static void
SetMenuTitle(
    MenuHandle menuHdl,		/* The menu we are setting the title of. */
    char *title)		/* The C string to set the title to. */
{
    int oldLength, newLength, oldHandleSize, dataLength;
    Ptr menuDataPtr;
 
    menuDataPtr = (Ptr) (*menuHdl)->menuData;

    if (strncmp(title, menuDataPtr + 1, menuDataPtr[0]) != 0) {    
    	newLength = strlen(title) + 1;
    	oldLength = menuDataPtr[0] + 1;
    	oldHandleSize = GetHandleSize((Handle) menuHdl);
    	dataLength = oldHandleSize - (sizeof(MenuInfo) - sizeof(Str255)) 
    		- oldLength;
    	if (newLength > oldLength) {
    	    SetHandleSize((Handle) menuHdl, oldHandleSize + (newLength 
    	    	    - oldLength));
    	    menuDataPtr = (Ptr) (*menuHdl)->menuData;
    	}
    
    	BlockMove(menuDataPtr + oldLength, menuDataPtr + newLength, 
    		dataLength);
    	BlockMove(title, menuDataPtr + 1, newLength - 1);
    	menuDataPtr[0] = newLength - 1;
    
    	if (newLength < oldLength) {
    	    SetHandleSize((Handle) menuHdl, oldHandleSize + (newLength 
    	    	    - oldLength));
    	}
    }
}

/*
 *----------------------------------------------------------------------
 *
 * TkpConfigureMenuEntry --
 *
 *	Processes configurations for menu entries.
 *
 * Results:
 *	Returns standard TCL result. If TCL_ERROR is returned, then
 *	interp->result contains an error message.
 *
 * Side effects:
 *	Configuration information get set for mePtr; old resources
 *	get freed, if any need it.
 *
 *----------------------------------------------------------------------
 */

int
TkpConfigureMenuEntry(
    Tcl_Interp *interp,			/* Used for error reporting. */
    Menu *menuPtr,			/* Information about whole menu. */
    register MenuEntry *mePtr,		/* Information about menu entry;  may
					 * or may not already have values for
					 * some fields. */
    int index)				/* Index of mePtr within menuPtr's
					 * entries. */
{
    MenuHandle macMenuHdl = (MenuHandle) menuPtr->platformData;
    Str255 itemText;

    /*
     * We have to do separators separately because SetMenuItemText
     * does not parse meta-characters.
     */
     
    if (mePtr->type == SEPARATOR_ENTRY) {
    	DeleteMenuItem(macMenuHdl, index + 1);
    	InsertMenuItem(macMenuHdl, SEPARATOR_TEXT, index);
    } else {
	GetEntryText(mePtr, itemText);
	SetMenuItemText(macMenuHdl, index + 1, itemText);
	
    	/*
    	 * Set enabling and disabling correctly.
    	 */

	if (mePtr->state == tkDisabledUid) {
	    DisableItem(macMenuHdl, index + 1);
	} else {
	    EnableItem(macMenuHdl, index + 1);
	}
    	
    	/*
    	 * Set the check mark for check entries and radio entries.
    	 */
	
	if (mePtr->type == CHECK_BUTTON_ENTRY
		|| mePtr->type == RADIO_BUTTON_ENTRY) {
	    CheckItem(macMenuHdl, index + 1, mePtr->entryFlags
		    & ENTRY_SELECTED);
	}
	
	if (mePtr->type == CASCADE_ENTRY) {
	    if ((mePtr->childMenuRefPtr != NULL) 
	    	    && (mePtr->childMenuRefPtr->menuPtr != NULL)) {
	    	MenuHandle childMenuHdl = (MenuHandle) mePtr->childMenuRefPtr
			->menuPtr->platformData;

		if (childMenuHdl == NULL) {
		    TkMenuCompleteIdleHandlers(
			    mePtr->childMenuRefPtr->menuPtr);
		    childMenuHdl = (MenuHandle) mePtr->childMenuRefPtr
			    ->menuPtr->platformData;
		}
		if (childMenuHdl != NULL) {
		    SetMenuCascade(mePtr->childMenuRefPtr->menuPtr);
	    	    SetItemMark(macMenuHdl, mePtr->index + 1,
			    (*childMenuHdl)->menuID);
	    	    SetItemCmd(macMenuHdl, mePtr->index + 1, CASCADE_CMD);
	    	    if (menuPtr->menuType == MENUBAR) {
	    	    	SetMenuTitle(childMenuHdl, mePtr->label);
	    	    }
	    	}
	    }
	} else {
	    SetMenuIndicator(mePtr);
	}
	
	/*
	 * We need to parse the accelerator string. If it has the strings
	 * for Command, Control, Shift or Option, we need to flag it
	 * so we can draw the symbols for it. We also need to precalcuate
	 * the position of the first real character we are drawing.
	 */
	
	if (0 == mePtr->accelLength) {
	    mePtr->platformEntryData = (PlatformEntryData) -1;
	} else {
	    char *accelString = mePtr->accel;
	    mePtr->entryFlags |= ~ENTRY_ACCEL_MASK;
	    
	    while (1) {
	    	if ((0 == strncasecmp("Control", accelString, 6))
	    		&& (('-' == accelString[6]) || ('+' == accelString[6]))) {
	  	    mePtr->entryFlags |= ENTRY_CONTROL_ACCEL;
	  	    accelString += 7;
	  	} else if ((0 == strncasecmp("Ctrl", accelString, 4))
	    		&& (('-' == accelString[4]) || ('+' == accelString[4]))) {
	  	    mePtr->entryFlags |= ENTRY_CONTROL_ACCEL;
	  	    accelString += 5;
	  	} else if ((0 == strncasecmp("Shift", accelString, 5))
	    		&& (('-' == accelString[5]) || ('+' == accelString[5]))) {
	  	    mePtr->entryFlags |= ENTRY_SHIFT_ACCEL;
	  	    accelString += 6;
	  	} else if ((0 == strncasecmp("Option", accelString, 6))
	    		&& (('-' == accelString[6]) || ('+' == accelString[6]))) {
	  	    mePtr->entryFlags |= ENTRY_OPTION_ACCEL;
	  	    accelString += 7;
	  	} else if ((0 == strncasecmp("Opt", accelString, 3))
	    		&& (('-' == accelString[3]) || ('+' == accelString[3]))) {
	  	    mePtr->entryFlags |= ENTRY_OPTION_ACCEL;
	  	    accelString += 4;
	  	} else if ((0 == strncasecmp("Command", accelString, 7))
	    		&& (('-' == accelString[7]) || ('+' == accelString[7]))) {
	  	    mePtr->entryFlags |= ENTRY_COMMAND_ACCEL;
	  	    accelString += 8;
	  	} else if ((0 == strncasecmp("Cmd", accelString, 3))
	    		&& (('-' == accelString[3]) || ('+' == accelString[3]))) {
	  	    mePtr->entryFlags |= ENTRY_COMMAND_ACCEL;
	  	    accelString += 4;
	  	} else if ((0 == strncasecmp("Alt", accelString, 3))
	    		&& (('-' == accelString[3]) || ('+' == accelString[3]))) {
	  	    mePtr->entryFlags |= ENTRY_OPTION_ACCEL;
	  	    accelString += 4;
	  	} else if ((0 == strncasecmp("Meta", accelString, 4))
	    		&& (('-' == accelString[4]) || ('+' == accelString[4]))) {
	  	    mePtr->entryFlags |= ENTRY_COMMAND_ACCEL;
	  	    accelString += 5;
	  	} else {
	  	    break;
	  	}
	    }
	    
	    if (ENTRY_COMMAND_ACCEL == (mePtr->entryFlags & ENTRY_ACCEL_MASK)) {
	    	SetItemCmd(macMenuHdl, mePtr->index+1, accelString[0]);
	    }
	    
	    mePtr->platformEntryData = (PlatformEntryData)
	    	    ((long) accelString - (long) mePtr->accel);
	}
    }
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * TkpPostMenu --
 *
 *	Posts a menu on the screen
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	The menu is posted and handled.
 *
 *----------------------------------------------------------------------
 */

int
TkpPostMenu(
    Tcl_Interp *interp,		/* The interpreter this menu lives in */
    Menu *menuPtr,		/* The menu we are posting */
    int x,			/* The global x-coordinate of the top, left-
    				 * hand corner of where the menu is supposed
    				 * to be posted. */
    int y)			/* The global y-coordinate */
{
    MenuHandle macMenuHdl = (MenuHandle) menuPtr->platformData;
    long popUpResult;
    int result;

    if (inPostMenu) {
        Tcl_AppendResult(interp,
		"Cannot call post menu while already posting menu",
		(char *) NULL);
    	result = TCL_ERROR;
    } else {
    	Window dummyWin;
    	unsigned int state;
    	int dummy, mouseX, mouseY;
    	short menuID;
	Window window;
	int oldWidth = menuPtr->totalWidth;
	Tk_Window parentWindow = Tk_Parent(menuPtr->tkwin);
    
    	inPostMenu++;
    	
    	result = TkPreprocessMenu(menuPtr);
    	if (result != TCL_OK) {
    	    inPostMenu--;
    	    return result;
    	}
    	
    	/*
    	 * The post commands could have deleted the menu, which means
    	 * we are dead and should go away.
    	 */
    	
    	if (menuPtr->tkwin == NULL) {
    	    inPostMenu--;
    	    return TCL_OK;
    	}
    	
	if (NULL == parentWindow) {
	    tearoffStruct.excludeRect.top = tearoffStruct.excludeRect.left
	    	    = tearoffStruct.excludeRect.bottom = tearoffStruct.excludeRect.right
	    	    = SHRT_MAX;
	} else {
	    int left, top;
	
	    Tk_GetRootCoords(parentWindow, &left, &top);
	    tearoffStruct.excludeRect.left = left;
	    tearoffStruct.excludeRect.top = top;
	    tearoffStruct.excludeRect.right = left + Tk_Width(parentWindow);
	    tearoffStruct.excludeRect.bottom = top + Tk_Height(parentWindow);
	    if (Tk_Class(parentWindow) == Tk_GetUid("Menubutton")) {
	    	TkWindow *parentWinPtr = (TkWindow *) parentWindow;
	    	TkMenuButton *mbPtr = 
	    		(TkMenuButton *) parentWinPtr->instanceData;
	    	int menuButtonWidth = Tk_Width(parentWindow)
	    		- 2 * (mbPtr->highlightWidth + mbPtr->borderWidth + 1);
	    	menuPtr->totalWidth = menuButtonWidth > menuPtr->totalWidth
	    		? menuButtonWidth : menuPtr->totalWidth;
	    }
	}
    	 
    	InsertMenu(macMenuHdl, -1);
    	RecursivelyInsertMenu(menuPtr);
    	CountMItems(macMenuHdl);
    	
	popUpResult = PopUpMenuSelect(macMenuHdl, y, x, menuPtr->active);
	menuPtr->totalWidth = oldWidth;
	RecursivelyDeleteMenu(menuPtr);
	DeleteMenu((*macMenuHdl)->menuID);
	
	/*
	 * Simulate the mouse up.
	 */
	 
	XQueryPointer(NULL, None, &dummyWin, &dummyWin, &mouseX,
	    &mouseY, &dummy, &dummy, &state);
	window = Tk_WindowId(menuPtr->tkwin);
	TkGenerateButtonEvent(mouseX, mouseY, window, state);
	
	/*
	 * Dispatch the command.
	 */
	 
	menuID = HiWord(popUpResult);
	if (menuID != 0) {
	    result = TkMacDispatchMenuEvent(menuID, LoWord(popUpResult));
	} else {
	    TkMacHandleTearoffMenu();
	    result = TCL_OK;
	}
	
	inPostMenu--;
    }
    return result;
}

/*
 *----------------------------------------------------------------------
 *
 * TkpMenuNewEntry --
 *
 *	Adds a pointer to a new menu entry structure with the platform-
 *	specific fields filled in. The Macintosh uses the
 *	platformEntryData field of the MenuEntry record to record
 *	the position of the first character in the accel string
 *	that is not a modifier.
 *
 * Results:
 *	Returns a pointer to the new entry.
 *
 * Side effects:
 *	Storage gets allocated.
 *
 *----------------------------------------------------------------------
 */

PlatformEntryData
TkpMenuNewEntry(
    MenuEntry *mePtr)		/* The menu we are adding an entry to */
{
    Menu *menuPtr = mePtr->menuPtr;

    /*
     * We have to do separators separately because we actually want
     * InsertMenuItem to parse the meta-characters. In all other cases
     * we do not want the meta-characters parsed.
     */

    MenuHandle macMenuHdl = (MenuHandle) menuPtr->platformData;
    Str255 itemText;

    if (mePtr->type == SEPARATOR_ENTRY) {
    	InsertMenuItem(macMenuHdl, SEPARATOR_TEXT, mePtr->index + 1);
    } else {

    	/*
    	 * We cannot call InsertMenuItem with the text directly because
    	 * the Menu Manager will parse out metacharacters. We want to
    	 * keep all of the slashes and so on.
     	 */

    	GetEntryText(mePtr, itemText);
    	InsertMenuItem(macMenuHdl, "\px", mePtr->index + 1);
    	SetMenuItemText(macMenuHdl, mePtr->index + 1, itemText);
    
    	SetMenuIndicator(mePtr);
    	
    }
    
    return NULL;
}

/*
 *----------------------------------------------------------------------
 *
 * 
 * DrawMenuBarWhenIdle --
 *
 *	Update the menu bar next time there is an idle event.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Menu bar is redrawn.
 *
 *----------------------------------------------------------------------
 */

static void
DrawMenuBarWhenIdle(
    ClientData clientData)	/* ignored here */
{
    DrawMenuBar();
    menuBarFlags &= ~MENUBAR_REDRAW_PENDING;
}


/*
 *----------------------------------------------------------------------
 *
 * RecursivelyInsertMenu --
 *
 *	Puts all of the cascades of this menu in the Mac hierarchical list.
 *
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	The menubar is changed.
 *
 *----------------------------------------------------------------------
 */

static void
RecursivelyInsertMenu(
    Menu *menuPtr)		/* All of the cascade items in this menu
    				 * will be inserted into the mac menubar. */
{
    int i;
    Menu *cascadeMenuPtr;
    MenuHandle macMenuHdl;
    
    for (i = 0; i < menuPtr->numEntries; i++) {
        if (menuPtr->entries[i]->type == CASCADE_ENTRY) {
            if ((menuPtr->entries[i]->childMenuRefPtr != NULL)
            	    && (menuPtr->entries[i]->childMenuRefPtr->menuPtr
		    != NULL)) {
            	cascadeMenuPtr = menuPtr->entries[i]->childMenuRefPtr->menuPtr;
	    	TkMenuCompleteIdleHandlers(cascadeMenuPtr);
	    	macMenuHdl = (MenuHandle) cascadeMenuPtr->platformData;
	    	InsertMenu(macMenuHdl, -1);
	    	RecursivelyInsertMenu(cascadeMenuPtr);
	    }
        }
    }
}

/*
 *----------------------------------------------------------------------
 *
 * RecursivelyDeleteMenu --
 *
 *	Puts all of the cascades of this menu in the Mac hierarchical list.
 *
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	The menubar is changed.
 *
 *----------------------------------------------------------------------
 */

static void
RecursivelyDeleteMenu(
    Menu *menuPtr)		/* All of the cascade items in this menu
    				 * will be inserted into the mac menubar. */
{
    int i;
    Menu *cascadeMenuPtr;
    MenuHandle macMenuHdl;
    
    for (i = 0; i < menuPtr->numEntries; i++) {
        if (menuPtr->entries[i]->type == CASCADE_ENTRY) {
            if ((menuPtr->entries[i]->childMenuRefPtr != NULL)
            	    && (menuPtr->entries[i]->childMenuRefPtr->menuPtr
		    != NULL)) {
            	cascadeMenuPtr = menuPtr->entries[i]->childMenuRefPtr->menuPtr;
	    	TkMenuCompleteIdleHandlers(cascadeMenuPtr);
	    	macMenuHdl = (MenuHandle) cascadeMenuPtr->platformData;
	    	DeleteMenu((*macMenuHdl)->menuID);
	    	RecursivelyInsertMenu(cascadeMenuPtr);
	    }
        }
    }
}

/*
 *----------------------------------------------------------------------
 *
 * SetDefaultMenubar --
 *
 *	Puts the Apple, File and Edit menus into the Macintosh menubar.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	The menubar is changed.
 *
 *----------------------------------------------------------------------
 */

static void
SetDefaultMenubar()
{
    if (currentMenuBarName != NULL) {
    	ckfree(currentMenuBarName);
    	currentMenuBarName = NULL;
    }
    currentMenuBarOwner = NULL;
    ClearMenuBar();
    InsertMenu(gAppleM, 0);
    InsertMenu(gFileM, 0);
    InsertMenu(gEditM, 0);
    if (!(menuBarFlags & MENUBAR_REDRAW_PENDING)) {
    	Tcl_DoWhenIdle(DrawMenuBarWhenIdle, (ClientData *) NULL);
    	menuBarFlags |= MENUBAR_REDRAW_PENDING;
    }
}

/*
 *----------------------------------------------------------------------
 *
 * TkpSetMainMenubar --
 *
 *	Puts the menu associated with a window into the menubar. Should
 *	only be called when the window is in front.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	The menubar is changed.
 *
 *----------------------------------------------------------------------
 */

void
TkpSetMainMenubar(
    Tcl_Interp *interp,		/* The interpreter of the application */
    Tk_Window tkwin,		/* The frame we are setting up */
    char *menuName)		/* The name of the menu to put in front.
    				 * If NULL, use the default menu bar.
    				 */
{
    MenuReferences *menuRefPtr;

    if (menuName != NULL && menuName[0] != NULL) {
        if (currentMenuBarOwner != tkwin 
        	&& (currentMenuBarName == NULL 
        	|| strcmp(menuName, currentMenuBarName) != 0)) {        
            
            if (currentMenuBarName != NULL) {
            	ckfree(currentMenuBarName);
            }
            
            currentMenuBarName = ckalloc(strlen(menuName) + 1);
            currentMenuBarOwner = tkwin;
            strcpy(currentMenuBarName, menuName);
		 
	    /*
	     * We need to find the clone of this menu that is the menubar.
	     * Once we do that, for every cascade in the menu, we need to 
	     * insert the Mac menu in the Mac menubar. Finally, we need
	     * to redraw the menubar.
	     */

	    menuRefPtr = TkFindMenuReferences(interp, menuName);	     
	    if (menuRefPtr != NULL) {
	    	Menu *menuPtr;
	    	Menu *cascadeMenuPtr;
	    	MenuHandle macMenuHdl;
	    	int i;
	        
	        menuPtr = menuRefPtr->menuPtr;
	        
	        while (menuPtr != NULL && menuPtr->menuType != MENUBAR) {
	            menuPtr = menuPtr->nextInstancePtr;
	        }
	        
	        if (menuPtr == NULL) {
	            SetDefaultMenubar();
	        } else {
		    ClearMenuBar();
	    	    InsertMenu(gAppleM, 0);

		    for (i = 0; i < menuPtr->numEntries; i++) {
		    	if (menuPtr->entries[i]->type == CASCADE_ENTRY) {
		    	    if ((menuPtr->entries[i]->childMenuRefPtr != NULL)
		    		    && menuPtr->entries[i]->childMenuRefPtr
				    ->menuPtr != NULL) {
		    	    	cascadeMenuPtr = menuPtr->entries[i]
				    	->childMenuRefPtr->menuPtr;
		    	    	TkMenuCompleteIdleHandlers(cascadeMenuPtr);
		    	    	macMenuHdl = (MenuHandle) cascadeMenuPtr
		    	    		->platformData;
    			    	DeleteMenu((*macMenuHdl)->menuID);
		    	    	InsertMenu(macMenuHdl, 0);
		    	    	RecursivelyInsertMenu(cascadeMenuPtr);
		    	    }
		    	}
		    }
		    if (!(menuBarFlags & MENUBAR_REDRAW_PENDING)) {
		    	Tcl_DoWhenIdle(DrawMenuBarWhenIdle, (ClientData *) NULL);
		    	menuBarFlags |= MENUBAR_REDRAW_PENDING;
		    }
		}
	    }
        }
    } else {
    	SetDefaultMenubar();
    }
}

/*
 *----------------------------------------------------------------------
 *
 * TkpSetWindowMenuBar --
 *
 *	Associates a given menu with a window. On the Mac, this means
 *	cloning all of the cascade menus underneath the menu.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	On Windows and UNIX, associates the platform menu with the
 *	platform window.
 *
 *----------------------------------------------------------------------
 */

void
TkpSetWindowMenuBar(
    Tk_Window tkwin,		/* The window we are setting the menu in */
    Menu *menuPtr)		/* The menu we are setting */
{
    /*
     * Nothing to do.
     */
}

/*
 *----------------------------------------------------------------------
 *
 * TkMacDispatchMenuEvent --
 *
 *	Given a menu id and an item, dispatches the command associated
 *	with it.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Commands get executed.
 *
 *----------------------------------------------------------------------
 */

int
TkMacDispatchMenuEvent(
    int menuID,		/* The menu id of the menu we are invoking */
    int index)		/* The one-based index of the item that was 
                         * selected. */
{
    int result = TCL_OK;
    if (menuID != 0) {
    	Tcl_HashEntry *commandEntryPtr = 
    		Tcl_FindHashEntry(&commandTable, (char *) menuID);
	Menu *menuPtr = (Menu *) Tcl_GetHashValue(commandEntryPtr);
	result = TkInvokeMenu(menuPtr->interp, menuPtr, index - 1);
    }
    return result;
}

/*
 *----------------------------------------------------------------------
 *
 * TkpGetMenuIndicatorGeometry --
 *
 *	Gets the width and height of the indicator area of a menu.
 *
 * Results:
 *	widthPtr and heightPtr are set.
 *
 * Side effects:
 *	None.
 *
 *----------------------------------------------------------------------
 */

void
TkpGetMenuIndicatorGeometry (
    Menu *menuPtr,			/* The menu we are drawing */
    MenuEntry *mePtr,			/* The entry we are measuring */
    int *widthPtr,			/* The resulting width */
    int *heightPtr,			/* The resulting height */
    Tk_Font tkfont,			/* Precalculated font */
    CONST Tk_FontMetrics *fmPtr)	/* Precalculated font metrics */
{
    char markChar;
    
    *heightPtr = fmPtr->linespace;
 
    markChar = (char) FindMarkCharacter(mePtr);
    *widthPtr = Tk_TextWidth(tkfont, &markChar, 1) + 2;
}

/*
 *----------------------------------------------------------------------
 *
 * TkpGetMenuAccelGeometry --
 *
 *	Gets the width and height of the accelerator area of a menu.
 *
 * Results:
 *	widthPtr and heightPtr are set.
 *
 * Side effects:
 *	None.
 *
 *----------------------------------------------------------------------
 */

void
TkpGetMenuAccelGeometry (
    Menu *menuPtr,			/* The menu we are measuring */
    MenuEntry *mePtr,			/* The entry we are measuring */
    int *widthPtr,			/* The resulting width */
    int *heightPtr,			/* The resulting height */
    Tk_Font tkfont,			/* The precalculated font */
    CONST Tk_FontMetrics *fmPtr)	/* The precalculated font metrics */
{
    *heightPtr = fmPtr->linespace;
    if (mePtr->type == CASCADE_ENTRY) {
        *widthPtr = SICN_HEIGHT;
    } else if (0 == mePtr->accelLength) {
    	*widthPtr = 0;
    } else {
    	
    	if (NULL == GetResource('SICN', SICN_RESOURCE_NUMBER)) {
    	    *widthPtr = Tk_TextWidth(tkfont, mePtr->accel,
    	    	    mePtr->accelLength);
    	} else {    	
    	    int length = (int) mePtr->platformEntryData;
    	    *widthPtr = 0;
    	    if (mePtr->entryFlags & ENTRY_CONTROL_ACCEL) {
    	    	*widthPtr += CONTROL_ICON_WIDTH;
    	    }
    	    if (mePtr->entryFlags & ENTRY_SHIFT_ACCEL) {
    	    	*widthPtr += SHIFT_ICON_WIDTH;
    	    }
    	    if (mePtr->entryFlags & ENTRY_OPTION_ACCEL) {
    	    	*widthPtr += OPTION_ICON_WIDTH;
    	    }
    	    if (mePtr->entryFlags & ENTRY_COMMAND_ACCEL) {
    	    	*widthPtr += COMMAND_ICON_WIDTH;
    	    }
    	    if (1 == length) {
    	    	*widthPtr = COMMAND_ICON_WIDTH;
    	    } else {
    	    	*widthPtr += Tk_TextWidth(tkfont, mePtr->accel 
    		    	+ length, mePtr->accelLength - length);
    	    }
    	}
    }
    *widthPtr += MARGIN_WIDTH;
}

/*
 *----------------------------------------------------------------------
 *
 * TkpGetTearoffEntryGeometry --
 *
 *	Gets the width and height of of a tearoff entry.
 *
 * Results:
 *	widthPtr and heightPtr are set.
 *
 * Side effects:
 *	None.
 *
 *----------------------------------------------------------------------
 */

void
TkpGetTearoffEntryGeometry (
    Menu *menuPtr,			/* The menu we are drawing */
    MenuEntry *mePtr,			/* The entry we are measuring */
    int *widthPtr,			/* The resulting width */
    int *heightPtr,			/* The resulting height */
    Tk_Font tkfont,			/* The precalculated font */
    CONST Tk_FontMetrics *fmPtr)	/* The precalculated font metrics */
{
    *widthPtr = *heightPtr = 0;
}

/*
 *----------------------------------------------------------------------
 *
 * TkpGetMenuSeparatorGeometry --
 *
 *	Gets the width and height of menu separator.
 *
 * Results:
 *	widthPtr and heightPtr are set.
 *
 * Side effects:
 *	None.
 *
 *----------------------------------------------------------------------
 */

void
TkpGetMenuSeparatorGeometry(
    Menu *menuPtr,			/* The menu we are drawing */
    MenuEntry *mePtr,			/* The entry we are measuring */
    int *widthPtr,			/* The resulting width */
    int *heightPtr,			/* The resulting height */
    Tk_Font tkfont,			/* The precalculated font */
    CONST Tk_FontMetrics *fmPtr)	/* The precalcualted font metrics */
{
    *widthPtr = 0;
    *heightPtr = fmPtr->linespace;
}

/*
 *----------------------------------------------------------------------
 *
 * TkpDrawMenuEntryIndicator --
 *
 *	This procedure draws the indicator part of a menu.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Commands are output to X to display the menu in its
 *	current mode.
 *
 *----------------------------------------------------------------------
 */

void
TkpDrawMenuEntryIndicator(
    Menu *menuPtr,			/* The menu we are drawing */
    MenuEntry *mePtr,			/* The entry we are drawing */
    Drawable d,				/* The drawable we are drawing */
    GC gc,				/* The GC we are drawing with */
    GC indicatorGC,			/* The GC to use for the indicator */
    Tk_Font tkfont,			/* The precalculated font */
    CONST Tk_FontMetrics *fmPtr,	/* The precalculated font metrics */
    int x,				/* topleft hand corner of entry */
    int y,				/* topleft hand corner of entry */
    int width,				/* width of entry */
    int height)				/* height of entry */
{
    if ((mePtr->type == CHECK_BUTTON_ENTRY || 
    	    mePtr->type == RADIO_BUTTON_ENTRY) 
    	    && mePtr->entryFlags & ENTRY_SELECTED) {
	int baseline;
	short markShort;
	char markChar;
    
    	baseline = y + (height + fmPtr->ascent - fmPtr->descent) / 2;
    	GetItemMark((MenuHandle) menuPtr->platformData,
    		mePtr->index + 1, &markShort);
        if (markShort != 0) {
            markChar = (char) markShort;
            Tk_DrawChars(menuPtr->display, d, gc, tkfont, &markChar, 1,
            	    x + 2, baseline);
	}
    }    
}

/*
 *----------------------------------------------------------------------
 *
 * DrawSICN --
 *
 *	Given a resource id and an index, loads the appropriate SICN
 *	and draws it into a given drawable using the given gc.
 *
 * Results:
 *	Returns 1 if the SICN was found, 0 if not found.
 *
 * Side effects:
 *	Commands are output to X to display the menu in its
 *	current mode.
 *
 *----------------------------------------------------------------------
 */
static int
DrawSICN(
    int resourceID,		    /* The resource # of the SICN table */
    int index,			    /* The index into the SICN table of the
    				     * icon we want. */
    Drawable d,			    /* What we are drawing into */
    GC gc,			    /* The GC to draw with */
    int x,			    /* The left hand coord of the SICN */
    int y)			    /* The top coord of the SICN */
{
    Handle sicnHandle = (Handle) GetResource('SICN', SICN_RESOURCE_NUMBER);
    
    if (NULL == sicnHandle) {
    	return 0;
    } else {
    	BitMap sicnBitmap;
	Rect destRect;
	CGrafPtr saveWorld;
	GDHandle saveDevice;
	GWorldPtr destPort;
	BitMapPtr destBitMap;

	HLock(sicnHandle);
	destPort = TkMacGetDrawablePort(d);
	GetGWorld(&saveWorld, &saveDevice);
	SetGWorld(destPort, NULL);
	TkMacSetUpClippingRgn(d);
	TkMacSetUpGraphicsPort(gc);

	SetRect(&destRect, x, y, x + SICN_HEIGHT, y + SICN_HEIGHT);
	sicnBitmap.baseAddr = (Ptr) (*sicnHandle) + index * SICN_HEIGHT
	    * SICN_ROWS;
	sicnBitmap.rowBytes = SICN_ROWS;
	SetRect(&sicnBitmap.bounds, 0, 0, 16, 16);
	destBitMap = &((GrafPtr) destPort)->portBits;
	CopyBits(&sicnBitmap, destBitMap, &sicnBitmap.bounds, &destRect, 
	    destPort->txMode, NULL);
	HUnlock(sicnHandle);
	SetGWorld(saveWorld, saveDevice);    
    	return 1;
    }
}

/*
 *----------------------------------------------------------------------
 *
 * TkpDrawMenuEntryAccelerator --
 *
 *	This procedure draws the accelerator part of a menu. We
 *	need to decide what to draw here. Should we replace strings
 *	like "Control", "Command", etc?
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Commands are output to X to display the menu in its
 *	current mode.
 *
 *----------------------------------------------------------------------
 */

void
TkpDrawMenuEntryAccelerator(
    Menu *menuPtr,		    /* The menu we are drawing */
    MenuEntry *mePtr,		    /* The entry we are drawing */
    Drawable d,			    /* The drawable we are drawing in */
    GC gc,			    /* The gc to draw into */
    Tk_Font tkfont,		    /* The precalculated font */
    CONST Tk_FontMetrics *fmPtr,    /* The precalculated font metrics */
    Tk_3DBorder activeBorder,	    /* border for menu background */
    int x,			    /* The left side of the entry */
    int y,			    /* The top of the entry */
    int width,			    /* The width of the entry */
    int height,			    /* The height of the entry */
    int drawArrow)		    /* Whether or not to draw cascade arrow */
{
    if (mePtr->type == CASCADE_ENTRY) {
    	if (0 == DrawSICN(SICN_RESOURCE_NUMBER, CASCADE_ARROW, d, gc,
    		x + width - SICN_HEIGHT, (y + (height / 2))
    		- (SICN_HEIGHT / 2))) {
	    XPoint points[3];
	    Tk_Window tkwin = menuPtr->tkwin;

	    if (mePtr->type == CASCADE_ENTRY) {
		points[0].x = width - menuPtr->activeBorderWidth
			- MARGIN_WIDTH - CASCADE_ARROW_WIDTH;
		points[0].y = y + (height - CASCADE_ARROW_HEIGHT)/2;
		points[1].x = points[0].x;
		points[1].y = points[0].y + CASCADE_ARROW_HEIGHT;
		points[2].x = points[0].x + CASCADE_ARROW_WIDTH;
		points[2].y = points[0].y + CASCADE_ARROW_HEIGHT/2;
		Tk_Fill3DPolygon(menuPtr->tkwin, d, activeBorder, points, 
			3, DECORATION_BORDER_WIDTH, TK_RELIEF_FLAT);
	    }
	}
    } else if (mePtr->accelLength != 0) {
    	int leftEdge = x + width - MARGIN_WIDTH;
    	int baseline = y + (height + fmPtr->ascent - fmPtr->descent) / 2;

	if (NULL == GetResource('SICN', SICN_RESOURCE_NUMBER)) {
	    leftEdge -= Tk_TextWidth(tkfont, mePtr->accel,
    	    	    mePtr->accelLength);
	    Tk_DrawChars(menuPtr->display, d, gc, tkfont, mePtr->accel,
	    	    mePtr->accelLength, leftEdge, baseline);
	} else {
	    int length = mePtr->accelLength - (int) mePtr->platformEntryData;
	    
	    if (length == 1) {
	    	leftEdge -= COMMAND_ICON_WIDTH;
	    } else {
	    	leftEdge -= Tk_TextWidth(tkfont, mePtr->accel
	    	    	+ (int) mePtr->platformEntryData,
	    	    	mePtr->accelLength - (int) mePtr->platformEntryData);
	    }
	    Tk_DrawChars(menuPtr->display, d, gc, tkfont, mePtr->accel 
		    + (int) mePtr->platformEntryData, mePtr->accelLength
		    - (int) mePtr->platformEntryData, leftEdge, baseline);

	    if (mePtr->entryFlags & ENTRY_COMMAND_ACCEL) {
	    	leftEdge -= COMMAND_ICON_WIDTH;
	    	DrawSICN(SICN_RESOURCE_NUMBER, COMMAND_ICON, d, gc,
	    		leftEdge, (y + (height / 2)) - (SICN_HEIGHT / 2));
	    }

	    if (mePtr->entryFlags & ENTRY_OPTION_ACCEL) {
	    	leftEdge -= OPTION_ICON_WIDTH;
	    	DrawSICN(SICN_RESOURCE_NUMBER, OPTION_ICON, d, gc,
	    		leftEdge, (y + (height / 2)) - (SICN_HEIGHT / 2));
	    }

	    if (mePtr->entryFlags & ENTRY_SHIFT_ACCEL) {
	    	leftEdge -= SHIFT_ICON_WIDTH;
	    	DrawSICN(SICN_RESOURCE_NUMBER, SHIFT_ICON, d, gc,
	    		leftEdge, (y + (height / 2)) - (SICN_HEIGHT / 2));
	    }

	    if (mePtr->entryFlags & ENTRY_CONTROL_ACCEL) {
	    	leftEdge -= CONTROL_ICON_WIDTH;
	    	DrawSICN(SICN_RESOURCE_NUMBER, CONTROL_ICON, d, gc,
	    		leftEdge, (y + (height / 2)) - (SICN_HEIGHT / 2));
	    }
    	}
    }
}

/*
 *----------------------------------------------------------------------
 *
 * TkpDrawTearoffEntry --
 *
 *	This draws a tearoff entry. On the Mac, these are zero-sized
 *	and are not drawn.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Commands are output to X to display the menu in its
 *	current mode.
 *
 *----------------------------------------------------------------------
 */

void
TkpDrawTearoffEntry(
    Menu *menuPtr,			/* The menu we are drawing */
    MenuEntry *mePtr,			/* The entry we are going to draw */
    Drawable d,				/* What we are drawing into */
    GC gc,				/* The gc to draw into */
    Tk_Font tkfont,			/* The precalculated font */
    CONST Tk_FontMetrics *fmPtr,	/* The precalculated font metrics */
    int x,				/* The left edge of the entry */
    int y,				/* The top edge of the entry */
    int width,				/* The width of the entry */
    int height)				/* The height of the entry */
{
    /*
     * Nothing to do
     */
}

/*
 *----------------------------------------------------------------------
 *
 * TkpDrawMenuSeparator --
 *
 *	The menu separator is drawn.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Commands are output to X to display the menu in its
 *	current mode.
 *
 *----------------------------------------------------------------------
 */

void
TkpDrawMenuSeparator(
    Menu *menuPtr,			/* The menu we are drawing */
    MenuEntry *mePtr,			/* The entry we are drawing */
    Drawable d,				/* The drawable we are drawing into */
    GC gc,				/* The gc we are drawing with */
    Tk_Font tkfont,			/* The precalculated font */
    CONST Tk_FontMetrics *fmPtr,	/* The precalculated font metrics */
    int x,				/* left coordinate of entry */
    int y,				/* top coordinate of entry */
    int width,				/* width of entry */
    int height)				/* height of entry */
{
    CGrafPtr saveWorld;
    GDHandle saveDevice;
    GWorldPtr destPort;
   
    destPort = TkMacGetDrawablePort(d);
    GetGWorld(&saveWorld, &saveDevice);
    SetGWorld(destPort, NULL);
    TkMacSetUpClippingRgn(d);
    
    /*
     * We don't want to use the text GC for drawing the separator. It
     * needs to be the same color as disabled items.
     */
    
    TkMacSetUpGraphicsPort(mePtr->disabledGC != None ? mePtr->disabledGC
    	    : menuPtr->disabledGC);
    
    MoveTo(x, y + (height / 2));
    Line(width, 0);
    
    SetGWorld(saveWorld, saveDevice);
}

/*
 *----------------------------------------------------------------------
 *
 * TkpDrawMenuUnderline --
 *
 *	On appropriate platforms, draw the underline character for the
 *	menu.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Commands are output to X to display the menu in its
 *	current mode.
 *
 *----------------------------------------------------------------------
 */

void
TkpDrawMenuUnderline(
    Menu *menuPtr,			/* The menu to draw into */
    MenuEntry *mePtr,			/* The entry we are drawing */
    Drawable d,				/* What we are drawing into */
    GC gc,				/* The gc to draw into */
    Tk_Font tkfont,			/* The precalculated font */
    CONST Tk_FontMetrics *fmPtr,	/* The precalculated font metrics */
    int x,				/* The left edge of the entry */
    int y,				/* The top edge of the entry */
    int width,				/* The width of the entry */
    int height)				/* The height of the entry */
{
    /*
     * Nothing to do.
     */
}

/*
 *----------------------------------------------------------------------
 *
 * MenuDefProc --
 *
 *	This routine is the MDEF handler for Tk. It receives all messages
 *	for the menu and dispatches them.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	This routine causes menus to be drawn and will certainly allocate
 *	memory as a result. Also, the menu can scroll up and down, and
 *	various other interface actions can take place.
 *
 *----------------------------------------------------------------------
 */

static void
MenuDefProc(
    short message,			/* What action are we taking? */
    MenuHandle menu,			/* The menu we are working with */
    Rect *menuRectPtr,			/* A pointer to the rect for the
    					 * whole menu. */
    Point hitPt,			/* Where the mouse was clicked for
    					 * the appropriate messages. */
    short *whichItem,			/* Output result. Which item was
    					 * hit by the user? */
    TkMenuLowMemGlobals *globalsPtr)	/* The low mem globals we have
    					 * to change */
{
    Menu *menuPtr;
    Tcl_HashEntry *commandEntryPtr;
    GrafPtr windowMgrPort;
    Tk_Font tkfont;
    Tk_FontMetrics fontMetrics, entryMetrics;
    Tk_FontMetrics *fmPtr;
    MenuEntry *mePtr;
    int i;
    int maxMenuHeight, menuHeight;
    int oldItem;
    int newItem = -1;
    GDHandle device;
    Rect itemRect;
    short windowPart;
    WindowRef whichWindow;
    RGBColor bgColor;
    RGBColor fgColor;
    RGBColor origFgColor;
    PenState origPenState;
    Rect dragRect;
    Rect scratchRect = {-32768, -32768, 32767, 32767};
    RgnHandle oldClipRgn;
    
    HLock((Handle) menu);
    commandEntryPtr = Tcl_FindHashEntry(&commandTable,
    	    (char *) (*menu)->menuID);
    HUnlock((Handle) menu);
    menuPtr = (Menu *) Tcl_GetHashValue(commandEntryPtr);

    switch (message) {
    	case mSizeMsg:
	    GetWMgrPort(&windowMgrPort);
    	    maxMenuHeight = windowMgrPort->portRect.bottom
    	    	    - windowMgrPort->portRect.top
    	    	    - GetMBarHeight() - 14;
    	    menuHeight = 0;
    	    
    	    for (i = 0; i < menuPtr->numEntries; i++) {
    	    	if ((menuHeight + menuPtr->entries[i]->height) > maxMenuHeight) {
    	    	    break;
    	    	}
    	    	menuHeight += menuPtr->entries[i]->height;
    	    }
    	    (*menu)->menuWidth = menuPtr->totalWidth;
    	    (*menu)->menuHeight = menuHeight;    	    
    	    break;

    	case mDrawMsg:    	
	    Tk_GetFontMetrics(menuPtr->tkfont, &fontMetrics);    	    
    	    
    	    GetGWorld(&macMDEFDrawable.portPtr, &device);
    	    
    	    for (i = 0; i < menuPtr->numEntries; i++) {
    	    	if (menuPtr->entries[i]->tkfont == NULL) {
    	    	    fmPtr = &fontMetrics;
    	    	    tkfont = menuPtr->tkfont;
    	    	} else {
    	    	    tkfont = menuPtr->entries[i]->tkfont;
    	    	    Tk_GetFontMetrics(tkfont, &entryMetrics);
    	    	    fmPtr = &entryMetrics;
    	    	}
    	    	TkDrawMenuEntry(menuPtr->entries[i], (Drawable) &macMDEFDrawable,
    	    		tkfont, fmPtr, menuRectPtr->left, 
    	    		menuRectPtr->top + menuPtr->entries[i]->y,
    	    		menuPtr->totalWidth, menuPtr->entries[i]->height, 
    	    		0, 1);
     	    }
     	    globalsPtr->menuTop = menuRectPtr->top;
     	    globalsPtr->menuBottom = menuRectPtr->bottom;
    	    break;

    	case mChooseMsg:
 	    GetGWorld(&macMDEFDrawable.portPtr, &device);

	    /*
	     * Find out which item was hit. If it is the same as the old item,
	     * we don't need to do anything.
	     */

	    oldItem = *whichItem - 1;
	     
	    if (PtInRect(hitPt, menuRectPtr)) {
	    	for (i = 0; i < menuPtr->numEntries; i++) {
	    	    itemRect.left = menuRectPtr->left;
	    	    itemRect.top = menuRectPtr->top + menuPtr->entries[i]->y;
	    	    itemRect.right = itemRect.left + menuPtr->totalWidth;
	    	    itemRect.bottom = itemRect.top + menuPtr->entries[i]->height;
	    	    if (PtInRect(hitPt, &itemRect)) {
	    	        if ((menuPtr->entries[i]->type == SEPARATOR_ENTRY)
	    	        	|| (menuPtr->entries[i]->state
	    	         	== tkDisabledUid)) {
	    	            newItem = -1;
	    	        } else {
	    	            newItem = i;
	    	            if ((menuPtr->entries[i]->type == CASCADE_ENTRY) &&
	    	            	    (oldItem != newItem)) {
		    	    	globalsPtr->itemRect = itemRect;
		    	    }
	    	        }
	    	        break;
	    	    }
	    	}
	    }

	    if (oldItem != newItem) {
	        if (oldItem >= 0) {
		    mePtr = menuPtr->entries[oldItem];
		    tkfont = mePtr->tkfont ? mePtr->tkfont : menuPtr->tkfont;
		    Tk_GetFontMetrics(tkfont, &fontMetrics);
		    TkDrawMenuEntry(mePtr, (Drawable) &macMDEFDrawable,
			    tkfont, &fontMetrics, menuRectPtr->left,
		    	    menuRectPtr->top + mePtr->y,
		    	    menuPtr->totalWidth, mePtr->height, 0, 1);
		}
		if (newItem != -1) {
		    int oldActiveItem = menuPtr->active;
		    
		    TkActivateMenuEntry(menuPtr, newItem);
		    mePtr = menuPtr->entries[newItem];
		    tkfont = mePtr->tkfont ? mePtr->tkfont : menuPtr->tkfont;
		    Tk_GetFontMetrics(tkfont, &fontMetrics);
		    TkDrawMenuEntry(mePtr, (Drawable) &macMDEFDrawable,
		    	    tkfont, &fontMetrics, menuRectPtr->left,
		    	    menuRectPtr->top + mePtr->y,
		    	    menuPtr->totalWidth, mePtr->height, 0, 1);
		    TkActivateMenuEntry(menuPtr, oldActiveItem);
		}
	    	*whichItem = newItem + 1;
	    
	    }
	    globalsPtr->menuDisable = ((*menu)->menuID << 16) | (newItem + 1);
	    
	    if (menuPtr->tearOff) {
   	    	scratchRect = *menuRectPtr;
		if (tearoffStruct.menuPtr == NULL) {
   	    	    scratchRect.top -= 10;
   	    	    scratchRect.bottom += 10;
   	    	    scratchRect.left -= 10;
   	    	    scratchRect.right += 10;
   	    	}

	    	windowPart = FindWindow(hitPt, &whichWindow);
	    	if ((windowPart != inMenuBar) && (newItem == -1)
	    		&& (hitPt.v != 0) && (hitPt.h != 0)
	    		&& (!PtInRect(hitPt, &scratchRect))
	    		&& (!PtInRect(hitPt, &tearoffStruct.excludeRect))) {
	    	    long dummy;
	    	    
	    	    oldClipRgn = NewRgn();
	    	    GetClip(oldClipRgn);
   	    	    GetForeColor(&origFgColor);
   	    	    GetPenState(&origPenState);
   	    	    GetForeColor(&fgColor);
   	    	    GetBackColor(&bgColor);
   	    	    GetGray(device, &bgColor, &fgColor);
   	    	    RGBForeColor(&fgColor);
   	    	    SetRect(&scratchRect, -32768, -32768, 32767, 32767);
   	    	    ClipRect(&scratchRect);
   	    	    
    	            dragRect = *menuRectPtr;
		    tearoffStruct.menuPtr = menuPtr;

	            PenMode(srcXor);
   	    	    dragRect = *menuRectPtr;
	            OffsetRect(&dragRect, -dragRect.left, -dragRect.top);
	            OffsetRect(&dragRect, tearoffStruct.point.h,
	        	    tearoffStruct.point.v);
	       	    if ((dragRect.top != 0) && (dragRect.left != 0)) {
	            	FrameRect(&dragRect);
	            	Delay(1, &dummy);
		    	FrameRect(&dragRect);
		    }
		    tearoffStruct.point = hitPt;

	    	    SetClip(oldClipRgn);
	    	    DisposeRgn(oldClipRgn);
	    	    RGBForeColor(&origFgColor);
	    	    SetPenState(&origPenState);    
		} else {
		    tearoffStruct.menuPtr = NULL;
		    tearoffStruct.point.h = tearoffStruct.point.v = 0;
		}
	    } else {
	    	tearoffStruct.menuPtr = NULL;
	    	tearoffStruct.point.h = tearoffStruct.point.v = 0;
	    }
	    
	    break;

    	case mPopUpMsg:
	    oldItem = *whichItem - 1;
	    if ((oldItem < 0) || (oldItem >= menuPtr->numEntries)) {
	        oldItem = 0;
	    }
	    menuRectPtr->top = menuRectPtr->bottom = hitPt.h;
    	    menuRectPtr->left = hitPt.v;
    	    menuRectPtr->right = menuRectPtr->left + menuPtr->totalWidth;
    	    for (i = 0; i < menuPtr->numEntries; i++) {
    	    	menuRectPtr->bottom += menuPtr->entries[i]->height;
    	    }
	    *whichItem = menuRectPtr->top;
    	    break;
    }
}


/*
 *----------------------------------------------------------------------
 *
 * TkMacHandleTearoffMenu() --
 *
 *	This routine sees if the MDEF has set a menu and a mouse position
 *	for tearing off and makes a tearoff menu if it has.
 *
 * Results:
 *	menuPtr->interp will have the result of the tearoff command.
 *
 * Side effects:
 *	A new tearoff menu is created if it is supposed to be.
 *
 *----------------------------------------------------------------------
 */

void
TkMacHandleTearoffMenu(void)
{
    if (tearoffStruct.menuPtr != NULL) {
    	Tcl_DString tearoffCmdStr;
    	char intString[20];
    	short windowPart;
    	WindowRef whichWindow;
    	
    	windowPart = FindWindow(tearoffStruct.point, &whichWindow);
    	
    	if (windowPart != inMenuBar) {
    	    Tcl_DStringInit(&tearoffCmdStr);
    	    Tcl_DStringAppendElement(&tearoffCmdStr, "tkTearOffMenuAtPoint");
    	    Tcl_DStringAppendElement(&tearoffCmdStr, 
    		    Tk_PathName(tearoffStruct.menuPtr->tkwin));
	    sprintf(intString, "%d", tearoffStruct.point.h);
	    Tcl_DStringAppendElement(&tearoffCmdStr, intString);
	    sprintf(intString, "%d", tearoffStruct.point.v);
	    Tcl_DStringAppendElement(&tearoffCmdStr, intString);
	    Tcl_Eval(tearoffStruct.menuPtr->interp,
		    Tcl_DStringValue(&tearoffCmdStr));
	    Tcl_DStringFree(&tearoffCmdStr);
	    tearoffStruct.menuPtr = NULL;
	}
    }
}

/*
 *--------------------------------------------------------------
 *
 * TkpInitializeMenuBindings --
 *
 *	For every interp, initializes the bindings for Windows
 *	menus. Does nothing on Mac or XWindows.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	C-level bindings are setup for the interp which will
 *	handle Alt-key sequences for menus without beeping
 *	or interfering with user-defined Alt-key bindings.
 *
 *--------------------------------------------------------------
 */

void
TkpInitializeMenuBindings(interp, bindingTable)
    Tcl_Interp *interp;		    /* The interpreter to set. */
    Tk_BindingTable bindingTable;   /* The table to add to. */
{
    /*
     * Nothing to do.
     */
}

/*
 *--------------------------------------------------------------
 *
 * TkpComputeMenubarGeometry --
 *
 *	This procedure is invoked to recompute the size and
 *	layout of a menu that is a menubar clone.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Fields of menu entries are changed to reflect their
 *	current positions, and the size of the menu window
 *	itself may be changed.
 *
 *--------------------------------------------------------------
 */

void
TkpComputeMenubarGeometry(menuPtr)
    Menu *menuPtr;		/* Structure describing menu. */
{
    TkComputeStandardMenuGeometry(menuPtr);
}

/*
 *----------------------------------------------------------------------
 *
 * TkpMenuInit --
 *
 *	Initializes Mac-specific menu data.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Allcates a hash table.
 *
 *----------------------------------------------------------------------
 */

void
TkpMenuInit(void)
{
    /* 
     * We have to set up the MDEF. This is pretty slimy. The real MDEF
     * resource is 68K code. All this code does is call another procedure.
     * When the application in launched, a dummy value for the procedure
     * is compiled into the MDEF. We are going to replace that dummy
     * value with a routine descriptor. When the routine descriptor
     * is invoked, the globals and everything will be setup, and we
     * can do what we need. This will not work from 68K or CFM 68k
     * currently, so we will conditional compile this until we
     * figure it out. 
     */
    
#ifdef GENERATINGCFM
    Handle MDEFHandle = GetResource('MDEF', 591);
    Handle SICNHandle = GetResource('SICN', SICN_RESOURCE_NUMBER);
    if ((MDEFHandle != NULL) && (SICNHandle != NULL)) {
    	HLock(MDEFHandle);
    	menuDefProc = TkNewMenuDefProc(MenuDefProc);
    	memmove((void *) (((long) (*MDEFHandle)) + 0x24), &menuDefProc, 4);
    	HUnlock(MDEFHandle);
    }
#endif

    lastMenuID = 256;
    Tcl_InitHashTable(&commandTable, TCL_ONE_WORD_KEYS);
    currentMenuBarOwner = NULL;
    tearoffStruct.menuPtr = NULL;
}
